##########################
#  Performance Options   #
##########################
fast_math = True
use_openmp = False 

##########################
# Insert Debugging Code  #
##########################
debug = False
check_pyobj_types = False 

#########################
#  Verbose Printing     #
#########################
print_input_ir = False
print_command_elapsed_time = False
print_commands = False
print_function_source = False
print_module_source = False

# Generate a .c file or a .cpp? 
pure_c = True
delete_temp_files = True 
